@@ -1,24 +1,27 @@
 # Python SCALE Codec Library
 #
-# Copyright 2018-2020 openAware BV (NL).
-# This file is part of Polkascan.
+# Copyright 2018-2020 Stichting Polkascan (Polkascan Foundation).
 #
-# Polkascan is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
 #
-# Polkascan is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
+# http://www.apache.org/licenses/LICENSE-2.0
 #
-# You should have received a copy of the GNU General Public License
-# along with Polkascan. If not, see <http://www.gnu.org/licenses/>.
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 from datetime import datetime
+from hashlib import blake2b
+
+from scalecodec.utils.ss58 import ss58_decode_account_index
+
 from scalecodec.base import ScaleType, ScaleBytes
 from scalecodec.exceptions import InvalidScaleTypeValueException
+from scalecodec.utils.math import trailing_zeros, next_power_of_two
 
 
 class Compact(ScaleType):
@@ -95,8 +98,10 @@ class Compact(ScaleType):
                 raise ValueError('{} out of range'.format(value))
 
 
-# Example of specialized composite implementation for performance improvement
 class CompactU32(Compact):
+    """
+    Specialized composite implementation for performance improvement
+    """
 
     type_string = 'Compact<u32>'
 
@@ -209,11 +214,15 @@ class String(ScaleType):
 
 class HexBytes(ScaleType):
 
+    def __init__(self, *args, **kwargs):
+        self.length_obj = None
+        super().__init__(*args, **kwargs)
+
     def process(self):
 
-        length = self.process_type('Compact<u32>').value
+        self.length_obj = self.process_type('Compact<u32>')
 
-        return '0x{}'.format(self.get_next_bytes(length).hex())
+        return '0x{}'.format(self.get_next_bytes(self.length_obj.value).hex())
 
     def process_encode(self, value):
 
@@ -228,6 +237,15 @@ class HexBytes(ScaleType):
         return data
 
 
+class CallBytes(ScaleType):
+
+    def process(self):
+        raise NotImplementedError()
+
+    def process_encode(self, value):
+        return bytes.fromhex(value[2:])
+
+
 class U8(ScaleType):
 
     def process(self):
@@ -404,160 +422,6 @@ class H512(ScaleType):
         return ScaleBytes(value)
 
 
-class VecU8Length64(ScaleType):
-    type_string = '[u8; 64]'
-
-    def process(self):
-        return '0x{}'.format(self.get_next_bytes(64).hex())
-
-    def process_encode(self, value):
-        if value[0:2] != '0x' and len(value) == 130:
-            raise ValueError('Value should start with "0x" and should be 64 bytes long')
-        return ScaleBytes(value)
-
-
-class VecU8Length32(ScaleType):
-    type_string = '[u8; 32]'
-
-    def process(self):
-        return '0x{}'.format(self.get_next_bytes(32).hex())
-
-    def process_encode(self, value):
-        if value[0:2] != '0x' and len(value) == 66:
-            raise ValueError('Value should start with "0x" and should be 32 bytes long')
-        return ScaleBytes(value)
-
-
-class VecU8Length20(ScaleType):
-    type_string = '[u8; 20]'
-
-    def process(self):
-        return '0x{}'.format(self.get_next_bytes(20).hex())
-
-    def process_encode(self, value):
-        if value[0:2] != '0x' and len(value) == 42:
-            raise ValueError('Value should start with "0x" and should be 20 bytes long')
-        return ScaleBytes(value)
-
-
-class VecU8Length16(ScaleType):
-    type_string = '[u8; 16]'
-
-    def process(self):
-        value = self.get_next_bytes(16)
-        try:
-            return value.decode()
-        except UnicodeDecodeError:
-            return value.hex()
-
-    def process_encode(self, value):
-        if value[0:2] != '0x' and len(value) == 34:
-            raise ValueError('Value should start with "0x" and should be 16 bytes long')
-        return ScaleBytes(value)
-
-
-class VecU8Length12(ScaleType):
-    type_string = '[u8; 12]'
-
-    def process(self):
-        value = self.get_next_bytes(12)
-        try:
-            return value.decode()
-        except UnicodeDecodeError:
-            return value.hex()
-
-    def process_encode(self, value):
-        if value[0:2] != '0x' and len(value) == 26:
-            raise ValueError('Value should start with "0x" and should be 12 bytes long')
-        return ScaleBytes(value)
-
-
-class VecU8Length8(ScaleType):
-    type_string = '[u8; 8]'
-
-    def process(self):
-        value = self.get_next_bytes(8)
-        try:
-            return value.decode()
-        except UnicodeDecodeError:
-            return value.hex()
-
-    def process_encode(self, value):
-        if value[0:2] != '0x' and len(value) == 18:
-            raise ValueError('Value should start with "0x" and should be 8 bytes long')
-        return ScaleBytes(value)
-
-
-class VecU8Length4(ScaleType):
-    type_string = '[u8; 4]'
-
-    def process(self):
-        value = self.get_next_bytes(4)
-        try:
-            return value.decode()
-        except UnicodeDecodeError:
-            return value.hex()
-
-    def process_encode(self, value):
-        if value[0:2] != '0x' and len(value) == 10:
-            raise ValueError('Value should start with "0x" and should be 4 bytes long')
-        return ScaleBytes(value)
-
-
-class VecU8Length2(ScaleType):
-    type_string = '[u8; 2]'
-
-    def process(self):
-        value = self.get_next_bytes(2)
-        try:
-            return value.decode()
-        except UnicodeDecodeError:
-            return value.hex()
-
-    def process_encode(self, value):
-        if value[0:2] != '0x' and len(value) == 6:
-            raise ValueError('Value should start with "0x" and should be 2 bytes long')
-        return ScaleBytes(value)
-
-
-class VecH256Length3(ScaleType):
-    type_string = '[H256; 3]'
-
-    def process(self):
-        return [self.process_type('H256').value, self.process_type('H256').value, self.process_type('H256').value]
-
-    def process_encode(self, value):
-        if type(value) is not list:
-            raise ValueError("Provided value is not a list")
-
-        data = None
-
-        for element in value:
-            element_obj = self.get_decoder_class('H256', metadata=self.metadata)
-            data += element_obj.encode(element)
-
-        return data
-
-
-class VecU128Length3(ScaleType):
-    type_string = '[u128; 3]'
-
-    def process(self):
-        return [self.process_type('u128').value, self.process_type('u128').value, self.process_type('u128').value]
-
-    def process_encode(self, value):
-        if type(value) is not list:
-            raise ValueError("Provided value is not a list")
-
-        data = None
-
-        for element in value:
-            element_obj = self.get_decoder_class('u128', metadata=self.metadata)
-            data += element_obj.encode(element)
-
-        return data
-
-
 class Struct(ScaleType):
 
     def __init__(self, data, type_mapping=None, **kwargs):
@@ -570,7 +434,10 @@ class Struct(ScaleType):
     def process(self):
 
         result = {}
+
         for key, data_type in self.type_mapping:
+            if data_type is None:
+                data_type = 'Null'
             result[key] = self.process_type(data_type, metadata=self.metadata).value
 
         return result
@@ -635,18 +502,102 @@ class Set(ScaleType):
 
 
 class Era(ScaleType):
+    """
+    An Era represents a range of blocks in which a transaction is allowed to be
+    executed.
+
+    An Era may either be "immortal", in which case the transaction is always valid,
+    or "mortal", in which case the transaction has a defined start block and period
+    in which it is valid.
+    """
+
+    def __init__(self, data=None, sub_type=None, metadata=None):
+        self.period = None
+        self.phase = None
+        super().__init__(data, sub_type, metadata)
 
     def process(self):
 
         option_byte = self.get_next_bytes(1).hex()
         if option_byte == '00':
+            self.period = None
+            self.phase = None
             return option_byte
         else:
-            return option_byte + self.get_next_bytes(1).hex()
+            encoded = int(option_byte, base=16) + (int(self.get_next_bytes(1).hex(), base=16) << 8)
+            self.period = 2 << (encoded % (1 << 4))
+            quantize_factor = max(1, (self.period >> 12))
+            self.phase = (encoded >> 4) * quantize_factor
+            if self.period >= 4 and self.phase < self.period:
+                return (self.period, self.phase)
+            else:
+                raise ValueError('Invalid phase and period: {}, {}'.format(self.phase, self.period))
 
+    def _tuple_from_dict(self, value):
+        if 'period' not in value:
+            raise ValueError("Value missing required field 'period' in dict Era")
+        period = value['period']
 
-class EraIndex(U32):
-    pass
+        if 'phase' in value:
+            return (period, value['phase'])
+
+        # If phase not specified explicitly, let the user specify the current block,
+        # and calculate the phase from that.
+        if 'current' not in value:
+            raise ValueError("Dict Era must have one of the fields 'phase' or 'current'")
+
+        current = value['current']
+
+        # Period must be a power of two between 4 and 2**16
+        period = max(4, min(1 << 16, next_power_of_two(period)))
+        phase = current % period
+        quantize_factor = max(1, (period >> 12))
+        quantized_phase = (phase // quantize_factor) * quantize_factor
+
+        return (period, quantized_phase)
+
+    def process_encode(self, value):
+        if value == '00':
+            self.period = None
+            self.phase = None
+            return ScaleBytes('0x00')
+        if isinstance(value, dict):
+            value = self._tuple_from_dict(value)
+        if isinstance(value, tuple) and len(value) == 2:
+            period, phase = value
+            if not isinstance(phase, int) or not isinstance(period, int):
+                raise ValueError("Phase and period must be ints")
+            if phase > period:
+                raise ValueError("Phase must be less than period")
+            self.period = period
+            self.phase = phase
+            quantize_factor = max(period >> 12, 1)
+            encoded = min(15, max(1, trailing_zeros(period) - 1)) | ((phase // quantize_factor) << 4)
+            return ScaleBytes(encoded.to_bytes(length=2, byteorder='little', signed=False))
+
+        raise ValueError("Value must be the string '00' or tuple of two ints")
+
+    def is_immortal(self) -> bool:
+        """Returns true if the era is immortal, false if mortal."""
+        return self.period is None or self.phase is None
+
+    def birth(self, current: int) -> int:
+        """Gets the block number of the start of the era given, with `current`
+        as the reference block number for the era, normally included as part
+        of the transaction.
+        """
+        if self.is_immortal():
+            return 0
+        return (max(current, self.phase) - self.phase) // self.period * self.period + self.phase
+
+    def death(self, current: int) -> int:
+        """Gets the block number of the first block at which the era has ended.
+
+        If the era is immortal, 2**64 - 1 (the maximum unsigned 64-bit integer) is returned.
+        """
+        if self.is_immortal():
+            return 2**64 - 1
+        return self.birth(current) + self.period
 
 
 class Bool(ScaleType):
@@ -663,10 +614,6 @@ class Bool(ScaleType):
             raise ValueError("Value must be boolean")
 
 
-class Moment(U64):
-    pass
-
-
 class CompactMoment(CompactU32):
     type_string = 'Compact<Moment>'
 
@@ -798,48 +745,21 @@ class Linkage(Struct):
     )
 
 
-class AccountId(H256):
+class GenericAccountId(H256):
+
+    def __init__(self, data=None, sub_type=None, metadata=None):
+        self.ss58_address = None
+        super().__init__(data, sub_type, metadata)
 
     def process_encode(self, value):
         if value[0:2] != '0x' and len(value) == 47:
             from scalecodec.utils.ss58 import ss58_decode
+            self.ss58_address = value
             value = '0x{}'.format(ss58_decode(value))
         return super().process_encode(value)
 
 
-class AccountIndex(U32):
-    pass
-
-
-class ReferendumIndex(U32):
-    pass
-
-
-class PropIndex(U32):
-    pass
-
-
-class Vote(U8):
-    pass
-
-
-class SessionKey(H256):
-    pass
-
-
-class SessionIndex(U32):
-    pass
-
-
-class Balance(U128):
-    pass
-
-
-class ParaId(U32):
-    pass
-
-
-class Key(Bytes):
+class GenericAccountIndex(U32):
     pass
 
 
@@ -848,10 +768,6 @@ class KeyValue(Struct):
     type_mapping = (('key', 'Vec<u8>'), ('value', 'Vec<u8>'))
 
 
-class BalanceOf(Balance):
-    pass
-
-
 class NewAccountOutcome(CompactU32):
     type_string = 'NewAccountOutcome'
 
@@ -893,28 +809,13 @@ class Vec(ScaleType):
         return data
 
 
-class VecNextAuthority(Vec):
-    type_string = 'Vec<NextAuthority>'
-
-    def process(self):
-        element_count = self.process_type('Compact<u32>').value
-
-        result = []
-        for _ in range(0, element_count):
-            element = self.process_type('NextAuthority')
-            self.elements.append(element)
-            result.append(element.value)
-
-        return result
-
-# class BalanceTransferExtrinsic(Decoder):
-#
-#     type_string = '(Address,Compact<Balance>)'
-#
-#     type_mapping = {'to': 'Address', 'balance': 'Compact<Balance>'}
+class BitVec(Vec):
+    # TODO: A BitVec that represents an array of bits. The bits are however stored encoded. The difference between this
+    #  * and a normal Bytes would be that the length prefix indicates the number of bits encoded, not the bytes
+    pass
 
 
-class Address(ScaleType):
+class GenericAddress(ScaleType):
 
     def __init__(self, data, **kwargs):
         self.account_length = None
@@ -957,7 +858,7 @@ class Address(ScaleType):
                 value = '0x{}'.format(ss58_decode(value))
             else:
                 from scalecodec.utils.ss58 import ss58_decode_account_index
-                index_obj = AccountIndex()
+                index_obj = GenericAccountIndex()
                 value = index_obj.encode(ss58_decode_account_index(value))
 
         if type(value) == str and value[0:2] == '0x' and len(value) == 66:
@@ -976,7 +877,7 @@ class Address(ScaleType):
             return self.value
 
 
-class AccountIdAddress(Address):
+class AccountIdAddress(GenericAddress):
 
     def process(self):
         self.account_id = self.process_type('AccountId').value.replace('0x', '')
@@ -991,7 +892,7 @@ class AccountIdAddress(Address):
                 value = '0x{}'.format(ss58_decode(value))
             else:
                 from scalecodec.utils.ss58 import ss58_decode_account_index
-                index_obj = AccountIndex()
+                index_obj = GenericAccountIndex()
                 value = index_obj.encode(ss58_decode_account_index(value))
 
         if type(value) == str and value[0:2] == '0x' and len(value) == 66:
@@ -1004,7 +905,7 @@ class AccountIdAddress(Address):
             raise ValueError('Value is in unsupported format, expected 32 bytes hex-string for AccountIds or int for AccountIndex')
 
 
-class RawAddress(Address):
+class RawAddress(GenericAddress):
     pass
 
 
@@ -1044,6 +945,10 @@ class Enum(ScaleType):
     def process_encode(self, value):
         if self.type_mapping:
 
+            if type(value) == str:
+                # Convert simple enum values
+                value = {value: None}
+
             if type(value) != dict:
                 raise ValueError("Value must be a dict when type_mapping is set, not '{}'".format(value))
 
@@ -1067,6 +972,14 @@ class Enum(ScaleType):
 
             raise ValueError("Value '{}' not present in value list of this enum".format(value))
 
+    def get_enum_value(self):
+        if self.value:
+
+            if self.type_mapping:
+                return list(self.value.values())[0]
+            else:
+                return self.value_list[self.index]
+
 
 class Data(Enum):
     type_mapping = [
@@ -1145,70 +1058,6 @@ class Data(Enum):
             raise ValueError("Value '{}' not present in type_mapping of this enum".format(enum_key))
 
 
-class RewardDestination(Enum):
-
-    value_list = ['Staked', 'Stash', 'Controller']
-
-
-class StakingLedger(Struct):
-    type_string = 'StakingLedger<AccountId, BalanceOf, BlockNumber>'
-    type_mapping = (
-        ('stash', 'AccountId'),
-        ('total', 'Compact<Balance>'),
-        ('active', 'Compact<Balance>'),
-        ('unlocking', 'Vec<UnlockChunk<Balance>>'),
-    )
-
-
-class UnlockChunk(Struct):
-    type_string = 'UnlockChunk<Balance>'
-    type_mapping = (
-        ('value', 'Compact<Balance>'),
-        ('era', 'Compact<EraIndex>'),
-    )
-
-
-class Exposure(Struct):
-    type_string = 'Exposure<AccountId, BalanceOf>'
-    type_mapping = (
-        ('total', 'Compact<Balance>'),
-        ('own', 'Compact<Balance>'),
-        ('others', 'Vec<IndividualExposure>'),
-    )
-
-
-class IndividualExposure(Struct):
-    type_string = 'IndividualExposure<AccountId, Balance>'
-    type_mapping = (
-        ('who', 'AccountId'),
-        ('value', 'Compact<Balance>'),
-    )
-
-
-class BabeAuthorityWeight(U64):
-    pass
-
-
-class KeyTypeId(VecU8Length4):
-    pass
-
-
-class Points(U32):
-    pass
-
-
-class EraPoints(Struct):
-    type_mapping = (
-        ('total', 'Points'),
-        ('individual', 'Vec<Points>'),
-    )
-
-
-class VoteThreshold(Enum):
-
-    value_list = ['SuperMajorityApprove', 'SuperMajorityAgainst', 'SimpleMajority']
-
-
 class Null(ScaleType):
 
     def process(self):
@@ -1222,61 +1071,6 @@ class InherentOfflineReport(Null):
     pass
 
 
-class LockPeriods(U8):
-    pass
-
-
-class Hash(H256):
-    pass
-
-
-class VoteIndex(U32):
-    pass
-
-
-class ProposalIndex(U32):
-    pass
-
-
-class Permill(U32):
-    pass
-
-
-class Perbill(U32):
-    pass
-
-
-class ApprovalFlag(U32):
-    pass
-
-
-class SetIndex(U32):
-    pass
-
-
-class AuthorityId(AccountId):
-    pass
-
-
-class ValidatorId(AccountId):
-    pass
-
-
-class AuthorityWeight(U64):
-    pass
-
-
-class StoredPendingChange(Struct):
-    type_mapping = (
-        ('scheduled_at', 'u32'),
-        ('forced', 'u32'),
-    )
-
-
-class ReportIdOf(Hash):
-    pass
-
-
 class StorageHasher(Enum):
 
     value_list = ['Blake2_128', 'Blake2_256', 'Blake2_128Concat', 'Twox128', 'Twox256', 'Twox64Concat', 'Identity']
@@ -1303,22 +1097,11 @@ class StorageHasher(Enum):
         return self.index == 6
 
 
-class VoterInfo(Struct):
-    type_string = 'VoterInfo<Balance>'
-
-    type_mapping = (
-        ('last_active', 'VoteIndex'),
-        ('last_win', 'VoteIndex'),
-        ('pot', 'Balance'),
-        ('stake', 'Balance'),
-    )
-
-
-class Gas(U64):
+class LockPeriods(U8):
     pass
 
 
-class CodeHash(Hash):
+class SessionKey(H256):
     pass
 
 
@@ -1334,22 +1117,6 @@ class PrefabWasmModule(Struct):
     )
 
 
-class OpaqueNetworkState(Struct):
-
-    type_mapping = (
-        ('peerId', 'OpaquePeerId'),
-        ('externalAddresses', 'Vec<OpaqueMultiaddr>'),
-    )
-
-
-class OpaquePeerId(Bytes):
-    pass
-
-
-class OpaqueMultiaddr(Bytes):
-    pass
-
-
 class SessionKeysSubstrate(Struct):
 
     type_mapping = (
@@ -1358,6 +1125,7 @@ class SessionKeysSubstrate(Struct):
         ('im_online', 'AccountId'),
     )
 
+
 class LegacyKeys(Struct):
 
     type_mapping = (
@@ -1416,72 +1184,6 @@ class VecQueuedKeys(Vec):
         return result
 
 
-class EthereumAddress(ScaleType):
-
-    def process(self):
-        value = self.get_next_bytes(20)
-        return value.hex()
-
-    def process_encode(self, value):
-        if value[0:2] == '0x' and len(value) == 42:
-            return ScaleBytes(value)
-        else:
-            raise ValueError('Value should start with "0x" and must be 20 bytes long')
-
-
-class EcdsaSignature(ScaleType):
-
-    def process(self):
-        value = self.get_next_bytes(65)
-        return value.hex()
-
-    def process_encode(self, value):
-        if value[0:2] == '0x' and len(value) == 132:
-            return ScaleBytes(value)
-        else:
-            raise ValueError('Value should start with "0x" and must be 65 bytes long')
-
-
-class BalanceLock(Struct):
-    type_string = 'BalanceLock<Balance, BlockNumber>'
-
-    type_mapping = (
-        ('id', 'LockIdentifier'),
-        ('amount', 'Balance'),
-        ('until', 'U32'),
-        ('reasons', 'WithdrawReasons'),
-    )
-
-
-class Bidder(Enum):
-    type_string = 'Bidder<AccountId, ParaIdOf>'
-
-    value_list = ['NewBidder', 'ParaId']
-
-
-class BlockAttestations(Struct):
-
-    type_mapping = (
-        ('receipt', 'CandidateReceipt'),
-        ('valid', 'Vec<AccountId>'),
-        ('invalid', 'Vec<AccountId>'),
-    )
-
-
-class IncludedBlocks(Struct):
-
-    type_mapping = (
-        ('actualNumber', 'BlockNumber'),
-        ('session', 'SessionIndex'),
-        ('randomSeed', 'H256'),
-        ('activeParachains', 'Vec<ParaId>'),
-        ('paraBlocks', 'Vec<Hash>'),
-    )
-
-class HeadData(Bytes):
-    pass
-
-
 class Conviction(Enum):
     CONVICTION_MASK = 0b01111111
     DEFAULT_CONVICTION = 0b00000000
@@ -1489,400 +1191,21 @@ class Conviction(Enum):
     value_list = ['None', 'Locked1x', 'Locked2x', 'Locked3x', 'Locked4x', 'Locked5x', 'Locked6x']
 
 
-class EraRewards(Struct):
-
-    type_mapping = (
-        ('total', 'u32'),
-        ('rewards', 'Vec<u32>'),
-    )
-
-
-class SlashJournalEntry(Struct):
-    type_mapping = (
-        ('who', 'AccountId'),
-        ('amount', 'Balance'),
-        ('ownSlash', 'Balance'),
-    )
-
-
-class UpwardMessage(Struct):
-    type_mapping = (
-        ('origin', 'ParachainDispatchOrigin'),
-        ('data', 'Bytes'),
-    )
-
-
-class ParachainDispatchOrigin(Enum):
-    value_list = ['Signed', 'Parachain']
-
-
-class StoredState(Enum):
-    value_list = ['Live', 'PendingPause', 'Paused', 'PendingResume']
-
-
-class Votes(Struct):
-    type_mapping = (
-        ('index', 'ProposalIndex'),
-        ('threshold', 'MemberCount'),
-        ('ayes', 'Vec<AccountId>'),
-        ('nays', 'Vec<AccountId>'),
-    )
-
-# Edgeware types
-# TODO move to RuntimeConfiguration per network
-
-
-class IdentityType(Bytes):
-    pass
-
-
-class VoteType(Enum):
-
-    type_string = 'voting::VoteType'
-
-    value_list = ['Binary', 'MultiOption']
-
-
-class VoteOutcome(ScaleType):
+class GenericBlock(ScaleType):
+    # TODO implement generic block type
 
     def process(self):
-        return list(self.get_next_bytes(32))
+        raise NotImplementedError()
+
+    def process_encode(self, value):
+        raise NotImplementedError()
 
 
-class Identity(Bytes):
+class GenericVote(U8):
     pass
 
 
-class ProposalTitle(Bytes):
-    pass
-
-
-class ProposalContents(Bytes):
-    pass
-
-
-class ProposalStage(Enum):
-    value_list = ['PreVoting', 'Voting', 'Completed']
-
-
-class ProposalCategory(Enum):
-    value_list = ['Signaling']
-
-
-class VoteStage(Enum):
-    value_list = ['PreVoting', 'Commit', 'Voting', 'Completed']
-
-
-class TallyType(Enum):
-
-    type_string = 'voting::TallyType'
-
-    value_list = ['OnePerson', 'OneCoin']
-
-
-class Attestation(Bytes):
-    pass
-
-
-# Joystream types
-# TODO move to RuntimeConfiguration per network
-
-class ContentId(H256):
-    pass
-
-
-class MemberId(U64):
-    pass
-
-
-class PaidTermId(U64):
-    pass
-
-
-class SubscriptionId(U64):
-    pass
-
-
-class SchemaId(U64):
-    pass
-
-
-class DownloadSessionId(U64):
-    pass
-
-
-class UserInfo(Struct):
-
-    type_mapping = (
-        ('handle', 'Option<Vec<u8>>'),
-        ('avatar_uri', 'Option<Vec<u8>>'),
-        ('about', 'Option<Vec<u8>>')
-    )
-
-
-class Role(Enum):
-
-    value_list = ['Storage']
-
-
-class ContentVisibility(Enum):
-    value_list = ['Draft', 'Public']
-
-
-class ContentMetadata(Struct):
-    type_mapping = (
-        ('owner', 'AccountId'),
-        ('added_at', 'BlockAndTime'),
-        ('children_ids', 'Vec<ContentId>'),
-        ('visibility', 'ContentVisibility'),
-        ('schema', 'SchemaId'),
-        ('json', 'Vec<u8>'),
-
-    )
-
-
-class ContentMetadataUpdate(Struct):
-    type_mapping = (
-        ('children_ids', 'Option<Vec<ContentId>>'),
-        ('visibility', 'Option<ContentVisibility>'),
-        ('schema', 'Option<SchemaId>'),
-        ('json', 'Option<Vec<u8>>')
-    )
-
-
-class LiaisonJudgement(Enum):
-    value_list = ['Pending', 'Accepted', 'Rejected']
-
-
-class BlockAndTime(Struct):
-    type_mapping = (
-        ('block', 'BlockNumber'),
-        ('time', 'Moment')
-    )
-
-
-class DataObjectTypeId(U64):
-    type_string = "<T as DOTRTrait>::DataObjectTypeId"
-
-
-class DataObject(Struct):
-    type_mapping = (
-        ('owner', 'AccountId'),
-        ('added_at', 'BlockAndTime'),
-        ('type_id', 'DataObjectTypeId'),
-        ('size', 'u64'),
-        ('liaison', 'AccountId'),
-        ('liaison_judgement', 'LiaisonJudgement'),
-        ('ipfs_content_id', 'Bytes'),
-    )
-
-
-class DataObjectStorageRelationshipId(U64):
-    pass
-
-
-class IPNSIdentity(Bytes):
-    pass
-
-
-class AccountInfo(Struct):
-    type_string = 'AccountInfo<BlockNumber>'
-
-    type_mapping = (
-        ('identity', 'IPNSIdentity'),
-        ('expires_at', 'BlockNumber'),
-    )
-
-
-class DownloadState(Enum):
-    value_list = ['Started', 'Ended']
-
-
-class DownloadSession(Struct):
-
-    type_mapping = (
-        ('content_id', 'ContentId'),
-        ('consumer', 'AccountId'),
-        ('distributor', 'AccountId'),
-        ('initiated_at_block', 'BlockNumber'),
-        ('initiated_at_time', 'BlockNumber'),
-        ('state', 'DownloadState'),
-        ('transmitted_bytes', 'u64'),
-    )
-
-
-class Url(Bytes):
-    pass
-
-
-class EntryMethod(Enum):
-    value_list = ['Paid', 'Screening']
-
-
-class Profile(Struct):
-    type_mapping = (
-        ('id', 'MemberId'),
-        ('handle', 'Bytes'),
-        ('avatar_uri', 'Bytes'),
-        ('about', 'Bytes'),
-        ('registered_at_block', 'BlockNumber'),
-        ('registered_at_time', 'Moment'),
-        ('entry', 'EntryMethod'),
-        ('suspended', 'bool'),
-        ('subscription', 'Option<SubscriptionId>'),
-    )
-
-
-class PaidMembershipTerms(Struct):
-    type_mapping = (
-        ('id', 'PaidTermId'),
-        ('fee', 'BalanceOf'),
-        ('text', 'Bytes'),
-    )
-
-
-class ThreadId(U64):
-    pass
-
-
-class InputValidationLengthConstraint(Struct):
-    type_mapping = (
-        ('min', 'u16'),
-        ('max_min_diff', 'u16'),
-    )
-
-
-class BlockchainTimestamp(Struct):
-    type_string = 'BlockchainTimestamp<BlockNumber, Moment>'
-
-    type_mapping = (
-        ('block', 'BlockNumber'),
-        ('time', 'Moment'),
-    )
-
-
-class ModerationAction(Struct):
-    type_mapping = (
-        ('moderated_at', 'BlockchainTimestamp<BlockNumber, Moment>'),
-        ('moderator_id', 'AccountId'),
-        ('rationale', 'Vec<u8>'),
-    )
-
-
-class PostId(U64):
-    pass
-
-
-class PostTextChange(Struct):
-    type_string = 'PostTextChange<BlockNumber, Moment>'
-
-    type_mapping = (
-        ('expired_at', 'BlockchainTimestamp<BlockNumber, Moment>'),
-        ('text', 'Vec<u8>'),
-    )
-
-
-class Post(Struct):
-    type_string = 'Post<BlockNumber, Moment, AccountId>'
-
-    type_mapping = (
-        ('id', 'PostId'),
-        ('thread_id', 'ThreadId'),
-        ('nr_in_thread', 'u32'),
-        ('current_text', 'Vec<u8>'),
-        ('moderation', 'Option<ModerationAction<BlockNumber, Moment, AccountId>>'),
-        ('text_change_history', 'Vec<PostTextChange<BlockNumber, Moment>>'),
-        ('created_at', 'BlockchainTimestamp<BlockNumber, Moment>'),
-        ('author_id', 'AccountId'),
-
-    )
-
-
-class Thread(Struct):
-    type_string = 'Thread<BlockNumber, Moment, AccountId>'
-
-    type_mapping = (
-        ('id', 'ThreadId'),
-        ('title', 'Vec<u8>'),
-        ('category_id', 'CategoryId'),
-        ('nr_in_category', 'u32'),
-        ('moderation', 'Option<ModerationAction<BlockNumber, Moment, AccountId>>'),
-        ('num_unmoderated_posts', 'u32'),
-        ('num_moderated_posts', 'u32'),
-        ('author_id', 'AccountId'),
-        ('created_at', 'BlockchainTimestamp<BlockNumber, Moment>'),
-        ('author_id', 'AccountId'),
-    )
-
-
-class CategoryId(U64):
-    pass
-
-
-class ChildPositionInParentCategory(Struct):
-
-    type_mapping = (
-        ('parent_id', 'CategoryId'),
-        ('child_nr_in_parent_category', 'u32'),
-    )
-
-
-class Category(Struct):
-    type_string = 'Category<BlockNumber, Moment, AccountId>'
-
-    type_mapping = (
-        ('id', 'CategoryId'),
-        ('title', 'Vec<u8>'),
-        ('description', 'Vec<u8>'),
-        ('created_at', 'BlockchainTimestamp<BlockNumber, Moment>'),
-        ('deleted', 'bool'),
-        ('archived', 'bool'),
-        ('num_direct_subcategories', 'u32'),
-        ('num_direct_unmoderated_threads', 'u32'),
-        ('num_direct_moderated_threads', 'u32'),
-        ('position_in_parent_category', 'Option<ChildPositionInParentCategory>'),
-        ('moderator_id', 'AccountId'),
-    )
-
-
-class ProposalStatus(Enum):
-    value_list = ['Active', 'Cancelled', 'Expired', 'Approved', 'Rejected', 'Slashed']
-
-
-class VoteKind(Enum):
-    value_list = ['Abstain', 'Approve', 'Reject', 'Slash']
-
-
-class RuntimeUpgradeProposal(Struct):
-    type_string = 'RuntimeUpgradeProposal<AccountId, Balance, BlockNumber, Hash>'
-
-    type_mapping = (
-        ('id', 'u32'),
-        ('proposer', 'AccountId'),
-        ('stake', 'Balance'),
-        ('name', 'Vec<u8>'),
-        ('description', 'Vec<u8>'),
-        ('wasm_hash', 'Hash'),
-        ('proposed_at', 'BlockNumber'),
-        ('status', 'ProposalStatus'),
-    )
-
-
-class TallyResult(Struct):
-    type_string = 'TallyResult<BlockNumber>'
-
-    type_mapping = (
-        ('proposal_id', 'u32'),
-        ('abstentions', 'u32'),
-        ('approvals', 'u32'),
-        ('rejections', 'u32'),
-        ('slashes', 'u32'),
-        ('status', 'ProposalStatus'),
-        ('finalized_at', 'BlockNumber'),
-    )
-
-
-class Call(ScaleType):
+class GenericCall(ScaleType):
 
     type_string = "Box<Call>"
 
@@ -1918,6 +1241,10 @@ class Call(ScaleType):
         }
 
     def process_encode(self, value):
+
+        if type(value) is not dict:
+            raise TypeError("value must be of type dict to encode a GenericCall")
+
         # Check requirements
         if 'call_index' in value:
             self.call_index = value['call_index']
@@ -1950,5 +1277,128 @@ class Call(ScaleType):
 
                     arg_obj = self.get_decoder_class(arg.type, metadata=self.metadata)
                     data += arg_obj.encode(param_value)
-
         return data
+
+
+class OpaqueCall(Bytes):
+
+    def process_encode(self, value):
+        call_obj = self.get_decoder_class('Call', metadata=self.metadata)
+        return super().process_encode(str(call_obj.encode(value)))
+
+    def process(self):
+        value = super().process()
+        try:
+            call_obj = self.get_decoder_class(
+                type_string='Call',
+                data=ScaleBytes('0x{}'.format(self.raw_value)),
+                metadata=self.metadata
+            )
+
+            return call_obj.process()
+        except:
+            return value
+
+
+class MultiAccountId(GenericAccountId):
+
+    @classmethod
+    def create_from_account_list(cls, accounts, threshold):
+        from scalecodec.utils.ss58 import ss58_decode
+
+        account_ids = []
+        for account in accounts:
+            if account[0:2] != '0x':
+                account = '0x{}'.format(ss58_decode(account))
+            account_ids.append(account)
+
+        account_list_cls = cls.get_decoder_class('Vec<AccountId>')
+        account_list_data = account_list_cls.encode(sorted(account_ids))
+        threshold_data = cls.get_decoder_class("u16").encode(threshold)
+
+        multi_account_id = "0x{}".format(blake2b(
+            b"modlpy/utilisuba" + bytes(account_list_data.data) + bytes(threshold_data.data), digest_size=32
+        ).digest().hex())
+
+        multi_account_obj = cls()
+        multi_account_obj.encode(multi_account_id)
+
+        return multi_account_obj
+
+
+class FixedLengthArray(ScaleType):
+
+    element_count = 0
+
+    def process(self):
+
+        if self.element_count:
+            if self.sub_type == 'u8':
+                return '0x{}'.format(self.get_next_bytes(self.element_count).hex())
+            else:
+                result = []
+                for idx in range(self.element_count):
+                    result.append(self.process_type(self.sub_type).value)
+        else:
+            result = []
+
+        return result
+
+    def process_encode(self, value):
+        data = ScaleBytes(bytearray())
+
+        value = value or []
+
+        if self.sub_type == 'u8':
+            # u8 arrays are represented as hex-bytes (e.g. [u8; 3] as 0x123456)
+            if value[0:2] != '0x' or len(value[2:]) != self.element_count * 2:
+                raise ValueError('Value should start with "0x" and should be {} bytes long'.format(self.element_count))
+
+            return ScaleBytes(value)
+
+        else:
+
+            if not type(value) is list:
+                raise ValueError('Given value is not a list')
+
+            for element_value in value:
+                element_obj = self.get_decoder_class(self.sub_type, metadata=self.metadata)
+                data += element_obj.encode(element_value)
+
+            return data
+
+class Ticker(ScaleType):
+
+    type_string = 'Ticker'
+
+    def process(self):
+        value = self.get_next_bytes(12)
+        try:
+            return value.decode()
+        except UnicodeDecodeError:
+            return value.hex()
+
+    def process_encode(self, value):
+        if value[0:2] != '0x' and len(value) == 26:
+            raise ValueError('Value should start with "0x" and should be 12 bytes long')
+        return ScaleBytes(value)
+
+
+class GenericMultiAddress(Enum):
+    type_mapping = [
+        ["Id", "AccountId"],
+        ["Index", "Compact<AccountIndex>"],
+        ["Raw", "Bytes"],
+        ["Address32", "H256"],
+        ["Address20", "H160"],
+      ]
+
+    def process_encode(self, value):
+
+        if type(value) == str:
+            if len(value) <= 8:
+                value = {"Index": ss58_decode_account_index(value)}
+            else:
+                value = {"Id": value}
+
+        return super().process_encode(value)
